import logging
import numpy as np
from clease.montecarlo.observers import MCObserver
from clease.tools import add_file_extension

logger = logging.getLogger(__name__)


class EnergyEvolution(MCObserver):
    """Trace the evolution of energy."""

    name = "EnergyEvolution"

    def __init__(self, mc, ignore_reset=False):
        super().__init__()
        self.mc = mc
        self.energies = []
        self.mean_energies = []
        self.ignore_reset = ignore_reset

    def __call__(self, system_changes):
        """Append the current energy to the MC object.

        Parameters:

        system_changes: list
            System changes. See doc-string of
            `clease.montecarlo.observers.MCObserver`
        """
        self.energies.append(self.mc.current_energy)
        self.mean_energies.append(self.mc.mean_energy.mean)

    def reset(self):
        """Reset the history."""
        if self.ignore_reset:
            return
        self.energies = []

    def save(self, fname: str = "energy_evolution.csv") -> None:
        """Save the energy evolution in .csv file.

        :param fname: File name of .csv file. Adds extension if none is given.
        """
        full_fname = add_file_extension(fname, '.csv')
        np.savetxt(full_fname, self.energies, delimiter=",")
        logger.info("Energy evolution data saved to %s.", full_fname)



import logging
import numpy as np
from .regression import LinearRegression, Tikhonov

logger = logging.getLogger(__name__)

__all__ = ('SequentialClusterRidge',)


class SequentialClusterRidge(LinearRegression):
    """
    SequentialClusterRidge is a fit method that optimizes the LOOCV over the
    regularization parameter as well as the cluster support. The method
    adds features in the design matrix X (see `fit` method) by including
    column by column. For each set of columns it performs a fit to a logspaced
    set of regularization parameters. The returned coefficients are the one
    from the model that has the smallest LOOCV.

    Parameters:

    alpha_min: float
        Minimum value of the regularization parameter alpha

    alpha_max: float
        Maximum value of the regularization parameter alpha

    num_alpha: int
        Number of alpha values
    """

    def __init__(self, min_alpha=1e-10, max_alpha=10.0, num_alpha=20):
        super().__init__()
        self.min_alpha = min_alpha
        self.max_alpha = max_alpha
        self.num_alpha = num_alpha

    @staticmethod
    def _cv(X, coeff, y, l2scheme):
        """
        Calcualtes the cross validation score
        """
        pred = X.dot(coeff)
        prec = l2scheme.precision_matrix(X)
        dy = y - pred
        dy_loo = dy / (1 - np.diag(X.dot(prec).dot(X.T)))
        return np.sqrt(np.mean(dy_loo**2))

    @staticmethod
    def _print_summary(cvs, coeffs):
        """
        Prints a summary of the search
        """
        srt_idx = np.argsort(cvs)
        print("--------------------------------------------")
        print("       SUPPORT EXPANDING L2 SUMMARY         ")
        print("--------------------------------------------")
        for i in range(20):
            print(f"Num. coeff: {len(coeffs[srt_idx[i]]):9d} "
                  f"CV: {cvs[srt_idx[i]]:9.3f} meV/atom")
        print("--------------------------------------------")

    def fit(self, X, y):
        """
        Performs the fitting

        Parameters:

        X: np.ndarray
            Design matrix of size (N x M). During the CV optimization columns
            of X will be added one by one starting with a model consisting
            of the two first columns.

        y: np.ndarray
            Vector of length N
        """
        numFeat = X.shape[1]
        alphas = np.logspace(np.log10(self.min_alpha), np.log10(self.max_alpha), self.num_alpha)

        coeffs = []
        cvs = []
        for i in range(2, numFeat):
            for alpha in alphas:
                scheme = Tikhonov(alpha=alpha)
                design = X[:, :i]
                coeff = scheme.fit(design, y)
                cv = self._cv(design, coeff, y, scheme)
                cvs.append(cv)
                coeffs.append(coeff)

        best_cv = np.argmin(cvs)
        res = np.zeros(numFeat)
        best_coeff = coeffs[best_cv]
        res[:len(best_coeff)] = best_coeff
        self._print_summary(cvs, coeffs)
        return res



from kivy.uix.floatlayout import FloatLayout
from kivy.properties import ObjectProperty


class HelpMessagePopup(FloatLayout):
    close = ObjectProperty(None)

    def __init__(self, **kwargs):
        msg = kwargs.pop('msg')
        FloatLayout.__init__(self, **kwargs)
        self.ids.message.text = msg



from kivy.uix.widget import Widget
from kivy.properties import ListProperty, StringProperty


class LegendItem(Widget):
    icon_color = ListProperty([1, 0, 0])
    icon_size = ListProperty([1, 1])
    text_color = ListProperty([1, 1, 1])
    text = StringProperty('sample')



# pylint: disable=undefined-variable
from .concentration import *
from .settings import *
from .settings_bulk import *
from .settings_slab import *
from .utils import *
from . import settings

ADDITIONAL = ('settings',)
__all__ = (concentration.__all__ + settings.__all__ + settings_bulk.__all__ +
           settings_slab.__all__ + utils.__all__ + ADDITIONAL)



from typing import Sequence
import numpy as np
from clease.datastructures import SystemChange


class BiasPotential:
    """
    Potential that can be used to manipulate the states visited.
    If added to MC run, the value returned from this will be
    added to the energy.
    """

    def __call__(self, system_changes: Sequence[SystemChange]):
        """Calculate the bias potential.

        Parameter

        system_changes: list
            List of tuples describing the changes
            see :py:method:`cemc.mcmc.mc_observers.MCObserver`
        """
        raise NotImplementedError("Has to be implemented in child classes!")

    def initialize(self):
        """Initialize the bias potential (if applicable).
           This functions is called right before MC sampling
           starts.
        """

    # pylint: disable=no-self-use,unused-argument
    def calculate_from_scratch(self, atoms):
        """Calculate the bias potential from scratch.

        Parameter:

        atoms: Atoms
            Structure to evaluate the bias poential for
        """
        msg = "Class has not implemented a calculate_from_scratch method!"
        raise NotImplementedError(msg)

    def save(self, fname="pseudo_binary_free_energy.pkl"):
        """Save the computed bias potential to a file.

        Parameter:

        fname: str
            Filename where a serialized version of this object
            will be stored.
        """
        import dill
        with open(fname, 'wb') as outfile:
            dill.dump(self, outfile)
        print(f"Pseudo binary free energy bias potential written to " f"{fname}")

    @staticmethod
    def load(fname="bias_potential.pkl"):
        """
        Load a bias potential from pickle file.
        Assume that this file has been stored with the save method
        of this class

        Parameter:

        fname: str
            Filename of a serialized version of this object
        """
        import dill
        with open(fname, 'rb') as infile:
            obj = dill.load(infile)
        return obj


class SampledBiasPotential(BiasPotential):
    """
    Class for bias potentials that are sampled.

    Parameters:

    reac_crd: list or numpy array
        Array with reaction coordinates

    free_eng: list or numpy array
        Array with energies
    """

    def __init__(self, reac_crd=(), free_eng=()):
        from scipy.interpolate import interp1d
        self.reac_crd = np.array(reac_crd)
        self.free_eng = np.array(free_eng)
        self.bias_interp = interp1d(self.reac_crd, self.free_eng, fill_value="extrapolate")

    def fit_smoothed_curve(self, smooth_length=11, show=False):
        """Fit a smoothed curve to the data.

        Parameters:

        smooth_length: int
            Window length of the Savitzky-Golay filter. Has to be odd.

        show: bool
            If True, a plot of the resulting curve will be shown.
        """
        if smooth_length % 2 == 0:
            raise ValueError("smooth_length has to be an odd number!")
        from scipy.signal import savgol_filter
        smoothed = savgol_filter(self.free_eng, smooth_length, 3)

        # Update the interpolator
        from scipy.interpolate import interp1d
        self.bias_interp = interp1d(self.reac_crd, smoothed, fill_value="extrapolate")

        if show:
            self.show()

    def show(self):
        """Create a plot of the bias potential."""
        from matplotlib import pyplot as plt
        fig = plt.figure()
        ax = fig.add_subplot(1, 1, 1)
        ax.plot(self.reac_crd, self.free_eng, 'o', mfc="none")
        ax.plot(self.reac_crd, self.bias_interp(self.reac_crd))
        ax.set_xlabel("Reaction coordinate")
        ax.set_ylabel("Bias potential")
        return fig

    def __iadd__(self, other):
        """Add two bias potentials.

        Parameter:

        other: BiasPotential
            The other potential
        """
        from scipy.interpolate import interp1d
        self.free_eng += other.get(self.reac_crd)
        self.bias_interp = interp1d(self.reac_crd, self.free_eng, fill_value="extrapolate")
        return self

    def __add__(self, other):
        from copy import deepcopy
        new_obj = deepcopy(self)
        new_obj += other
        return new_obj


class PseudoBinaryFreeEnergyBias(SampledBiasPotential):
    """
    Class for a bias potential based on the free energy curve.
    It is intended for use when the reaction coordinate is the composition
    of one of the pseudobinary groups.
    NOTE: In the future this can in principle be extended to handle the
    case of arbitrary reaction paths

    Parameters:


    pseudo_bin_conc_init: PseudoBinaryConcInitializer or None
        Can be None if __call__ is not called (i.e.) for fitting a smoothed
        curve etc.
    reac_crd: list or numpy array
        Value of the reaction coordinate

    free_eng: list or numpy array
        Value of the free energy corresponding to the reac_crd
        array
    """

    def __init__(self, pseudo_bin_conc_init=None, reac_crd=(), free_eng=()):
        self._conc_init = pseudo_bin_conc_init
        super(PseudoBinaryFreeEnergyBias, self).__init__(reac_crd, free_eng)

    @property
    def conc_init(self):
        from cemc.mcmc import PseudoBinaryConcInitializer
        if not isinstance(self._conc_init, PseudoBinaryConcInitializer):
            raise TypeError("pseudo_bin_conc_init has to be of type "
                            "PseudoBinaryConcInitializer!")
        return self._conc_init

    @conc_init.setter
    def conc_init(self, init):
        from cemc.mcmc import PseudoBinaryConcInitializer
        if not isinstance(init, PseudoBinaryConcInitializer):
            raise TypeError("pseudo_bin_conc_init has to be of type "
                            "PseudoBinaryConcInitializer!")
        self._conc_init = init

    def __call__(self, system_changes: Sequence[SystemChange]):
        """Evaluate the bias potential.

        see :py:method:`cemc.mcmc.mc_observers_MCObserver.__call__`
        """
        # We can pass None  in this case
        # PseudoBinaryConcInitializer should track the atoms object itself
        q = self.conc_init.get(None)
        symb = self.conc_init.target_symb
        for change in system_changes:
            if change[1] == symb:
                q -= 1.0 / self.conc_init.num_per_unit
            elif change[2] == symb:
                q += 1.0 / self.conc_init.num_per_unit
        return self.bias_interp(q)

    def calculate_from_scratch(self, atoms):
        """Calculate the value from scratch."""
        q = self.conc_init.get(atoms)
        return self.bias_interp(q)

    def get(self, reac_crd):
        """Get the bias potential as a function of reaction coordinate.

        Parameter:

        reac_crd: float
            Reaction coordinate
        """
        return self.bias_interp(reac_crd)


class CovarianceBiasPotential(SampledBiasPotential):
    """
    Bias potential to be used together with
    :py:class:`cemc.mcmc.cov_reac_crd.CovarianceRangeConstraint`

    Parameters:

    cov_range: CovarianceRangeConstraint or None
        Range constraints, mainly used to calculating the
        value of the reaction coordinate.

    reac_crd: list or numpy array
        Reaction coordinates

    free_eng: list or numpy array
        Free energy at each reaction coordinate
    """

    def __init__(self, cov_range=None, reac_crd=(), free_eng=()):
        super(CovarianceBiasPotential, self).__init__(reac_crd, free_eng)
        self._cov_range = cov_range

    @property
    def cov_range(self):
        from cemc.mcmc import CovarianceRangeConstraint
        if not isinstance(self._cov_range, CovarianceRangeConstraint):
            raise TypeError("cov_range has to be of type CovarianceRangeConstraint")
        return self._cov_range

    @cov_range.setter
    def cov_range(self, obj):
        self._cov_range = obj

    def __call__(self, system_changes: Sequence[SystemChange]):
        """Get the value for the bias potential.

        Parameter:

        system_changes: list
            Proposed atom moves
        """
        reac_crd = self.cov_range.get_new_value(system_changes)
        return self.bias_interp(reac_crd)

    def calculate_from_scratch(self, atoms):
        """Calculate the value from scratch."""
        reac_crd = self.cov_range._cov_init.get(atoms, None)
        return self.bias_interp(reac_crd)

    def get(self, reac_crd):
        """Get the bias as a function reaction coordinate.abs

        Parameters:

        reac_crd: float
            Reaction coordinate
        """
        return self.bias_interp(reac_crd)



from clease.montecarlo.bias_potential import BiasPotential
from clease.montecarlo.montecarlo import Montecarlo
from clease.montecarlo.sgc_montecarlo import SGCMonteCarlo
from clease.montecarlo.gaussian_kernel_bias_potential import GaussianKernelBiasPotential
from clease.montecarlo.binned_bias_potential import BinnedBiasPotential
from clease.montecarlo.random_bias_with_memory import RandomBiasWithMemory
from clease.montecarlo.metadynamics_sampler import MetaDynamicsSampler
# pylint: disable=undefined-variable
from .barrier_models import *
from .kmc_events import *
from .kinetic_monte_carlo import *
from .trial_move_generator import *

ADDITIONAL = ('Montecarlo', 'SGCMonteCarlo', 'MetaDynamicsSampler', 'BinnedBiasPotential',
              'BiasPotential', 'GaussianKernelBiasPotential', 'RandomBiasWithMemory')
__all__ = (ADDITIONAL + barrier_models.__all__ + kinetic_monte_carlo.__all__ + kmc_events.__all__,
           trial_move_generator.__all__)
